#include <algorithm>
#include <iostream>
#include <sstream>
#include <rlog/rlog.h>

#include <assert.h>

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "Compress.hpp"
#include "FileRawCompressed.hpp"
#include "FileRawNormal.hpp"
#include "TransformTable.hpp"
#include "CompressedTable.hpp"

extern TransformTable	*g_TransformTable;
static CompressedTable	 g_CompressedTable;

/*
 * Output:
 * 	fh - Restored file header, valid only if true is returned.
 */
bool Compress::isCompressed(const char *name, FileHeader &fh)
{
	int fd;

	// Using File::open(name) because it only calls ::open
	// and if it fails, it calls open_force. Reference counts are
	// not changed...
	//
	fd = File::open(name);
	if (fd < 0)
		return false;
	
	fh.restore(fd, 0);

	::close(fd);

	return fh.isValid();
}

void Compress::createFileRaw(const char *name)
{
	bool		fl = false;
	FileHeader	fh;

	if (m_FileSize != 0)
	{
		fl = isCompressed(name, fh);
	}

	// Decide if we create instance of FileRaw for
	// compressed or uncompressed files.
	//
	if ((g_CompressedTable.isNativelyCompressed(name) == false) || (fl))
	{
		if (m_FileSize == 0)
		{
			fh.type = g_TransformTable->getDefaultIndex();
		}
		
		if ((m_FileSize == 0) || fl)
		{
			rDebug("C (%s), real/orig %ld/%ld bytes",
					name, (long int) m_FileSize, (long int) fh.size);
			m_FileRaw = new (std::nothrow) FileRawCompressed(fh, m_FileSize);
		}
	}

	if (!m_FileRaw)
	{
		rDebug("N (%s)", name);
		m_FileRaw = new (std::nothrow) FileRawNormal();
	}

	if (!m_FileRaw)
	{
		rError("No memory to allocate object of FileRaw interface");
		abort();
	}
}

Compress::Compress(const struct stat *st) :
	File (st),
	m_FileRaw (NULL),
	m_FileSize (st->st_size)
{
}

Compress::~Compress()
{
	delete m_FileRaw;
}

int Compress::unlink(const char *name)
{
	rDebug("%s", __FUNCTION__);

	delete m_FileRaw;
	
	m_FileSize = 0;
	m_FileRaw = NULL;

	return File::unlink(name);
}

int Compress::truncate(const char *name, off_t size)
{
	if (!m_FileRaw)
		createFileRaw(name);

	return m_FileRaw->truncate(name, size);
}

int Compress::getattr(const char *name, struct stat *st)
{
	if (!m_FileRaw)
		createFileRaw(name);

	return m_FileRaw->getattr(name, st);
}

int Compress::open(const char *name, int flags)
{
	int r;

	r = File::open(name, flags);

	if (!m_FileRaw)
		createFileRaw(name);

	if (m_refs == 1)
	{
		m_FileRaw->open(m_fd);
	}

	return r;
}

int Compress::release(const char *name)
{
	if ((m_refs == 1) && (m_FileRaw))
	{
		m_FileRaw->close();
	}
	
	return File::release(name);
}

ssize_t Compress::write(const char *buf, size_t size, off_t offset)
{
	assert (m_FileRaw);
	
	rDebug("Compress::write size: %d, offset: %lld", (unsigned int) size,
			(long long int) offset);

	return m_FileRaw->write(buf, size, offset);
}

ssize_t Compress::read(char *buf, size_t size, off_t offset)
{
	assert (m_FileRaw);

	rDebug("Compress::read size: %d, offset: %lld", (unsigned int) size,
			(long long int) offset);

	return m_FileRaw->read(buf, size, offset);
}

