#include <errno.h>
#include <limits.h>

#include <cassert>
#include <ext/functional>
#include <rlog/rlog.h>

#include "CompressMap.hpp"
#include "TransformCompressLZO.hpp"

using namespace std;

CompressMap::CompressMap(Transform *transform) :
	m_size (0),
	m_length (0),
	m_transform (transform)
{
}

void CompressMap::set(Transform *transform)
{
	m_transform = transform;
}

int CompressMap::store(char *buf, size_t len) const
{
	int				r;
	con_t::const_iterator		it;

	assert (len >= m_size * Block::size());

	for (it = m_map.begin(); it != m_map.end(); ++it)
	{
		// The order of the pointers to the 'Block' in vector must
		// be preserved...
		// 
		r = (*it)->store(buf, len);
		if (r == -1)
			return -1;

		buf += r;
		len -= r;
	}
	return 0;
}

off_t CompressMap::store(int fd, off_t to) const
{
	size_t			 len;
	char			*buf;
	Block			*bl;

	if (m_size == 0)
		return 0;

	// Create buffer to store the index.
	//
	len = Block::size() * m_size;
	buf = new (std::nothrow) char[len];
	if (!buf)
	{
		rError("No memory to allocate block of %d bytes",
				len);

		errno = -ENOMEM;
		return -1;
	}

	if (store(buf, len) == 1)
	{
		delete[] buf;

		return -1;
	}

	// Compress buffer with index and store it
	// to the disk. Reserve space for the Block describing
	// this compressed index.
	//
	bl = m_transform->store(fd, to + Block::size(),
				    to + Block::size(), buf, len);
	delete[] buf;

	if (!bl)
		return -1;

	// Store block describing compressed index.
	// 
	to = bl->store(fd, to);
	if (to == (off_t) -1)
		goto out;

	// Correct file position. We must add buffer's compressed
	// length...
	// 
	to += bl->clength;
out:
	delete bl;

	return to;
}

int CompressMap::restore(size_t count, const char *buf, size_t len)
{
	int	 r;
	Block	*bl;
	
	// Use 'count--', not '--count' !!
	// 
	while (count--)
	{
		bl = new (std::nothrow) Block();
		if (!bl)
		{
			rError("No memory to allocate object of Block class");

			errno = -ENOMEM;
			return -1;
		}

		r = bl->restore(buf, len);
		if (r == (off_t) -1)
		{
			delete bl;
			
			return -1;
		}
		
		buf += r;
		len -= r;
		
		put(bl);
	}
	return 0;
}

off_t CompressMap::restore(int fd, off_t from)
{
	char			*buf;
	size_t			 count;
	Block			 bl;

	// Delete old linear map before a new one is
	// restored from the disk.
	//
	flush();

	// Read block describing compressed index.
	// 
	from = bl.restore(fd, from);
	if (from == (off_t) -1)
		return -1;

	// Compute number of blocks in the index.
	//
	count = bl.length / Block::size();

	// There is no index (it's zero length file...)
	// 
	if (count == 0)
		return from;

	// Create buffer for uncompressed index.
	//
	buf = new (std::nothrow) char[bl.length];
	if (!buf)
	{
		rError("No memory to allocate block of %d bytes",
				bl.length);

		errno = -ENOMEM;
		return -1;
	}

	// Decompress index from file to buffer.
	//
	from = m_transform->restore(fd, &bl, from, buf, bl.length);
	if (from == (off_t) -1)
		goto out;

	// Retrieve blocks and store them in this
	// class.
	// 
	if (restore(count, buf, bl.length) == -1)
		from = -1;
out:	
	delete[] buf;
	
	return from;
}

void CompressMap::truncate(off_t length)
{
	if (length == 0)
	{
		flush();
	}
	else
	{
		// TODO: Function is not implemented, it is
		// not so important, because without implementing this,
		// index will grow until it is written to
		// the disk... But you must limit get() somehow...
		//
	}
}

void CompressMap::flush()
{
	con_t::iterator it = m_map.begin();

	// Deallocate elements from heap memory...
	//
	while (it != m_map.end())
	{
		delete *it;
		m_map.erase(it++);
	}

	m_size = 0;
	m_length = 0;
}

CompressMap::~CompressMap()
{
	flush();
}

/**
 * Prerequirement: 'it' must points to the highest level on that
 * offset, ie. that simple --it turns it to point to the lowest
 * level on previous offset.
 */
void CompressMap::prev(con_t::const_iterator &it) const
{
	con_t::const_iterator prev;
	
	if (--it == m_map.end())
		return;
	
	prev = it;
	
	while (--prev != m_map.end())
	{
		if ((*prev)->offset != (*it)->offset)
			break;

		it = prev;

		if (it == m_map.begin())
			break;
	}
}

void CompressMap::next(con_t::const_iterator &it) const
{
	off_t offset;
	
	if (it == m_map.end())
		return;
	
	offset = (*it)->offset;
	
	while (++it != m_map.end())
	{
		if (offset != (*it)->offset)
			break;
	}
}

/**
 * 
 * @param offset
 * @param bl
 * @param it
 *
 * @return
 */
off_t CompressMap::get_length(off_t			offset,
                            con_t::const_iterator	it) const
{
	Block	*bl = *it;
	Block	*nbl;

	assert (bl->offset + bl->length > offset);

	if (bl->offset > offset)
	{
		// Block starts higher than offset.
		//
		return 0;
	}

	// Iterate over offsets. We want only Block with
	// the highest level in that offset.
	//
	while (true)
	{
		next(it);

		if (it == m_map.end())
			break;
		
		nbl = *it;
		
		// Check if block 'nbl' overlaps block 'bl'.
		//
		if (nbl->offset > bl->offset + bl->length)
			break;

		if (nbl->offset <= offset)
			continue;
		
		if (bl->level < nbl->level)
		{
			// Yes, restrict length of the Block we found
			// earlier.
			//
			//                   nbl->offset
			//                   |--------------|
			//    bl->offset
			//    | offset
			//    | |
			//    |---bl->length---|
			//
			assert (bl->offset <= offset);
			assert (nbl->offset - offset > 0);

			return nbl->offset - offset;
		}
	}
	assert (bl->length - (offset - bl->offset) > 0);
	
	return bl->length - (offset - bl->offset);
}

/**
 * Function called only when there is no block before offset or
 * when no block before offset overlaps over it.
 *
 * @param it - iterator to vector containing blocks that start on
 *             offset or bigger.
 */
off_t CompressMap::get_simple(off_t			  offset,
                            con_t::const_iterator	  it,
                            Block			**bl) const
{
	// Block must point to the highest level on that offset.
	// 
	if (it != m_map.end())
	{
		Block *block = *it;

		assert (block->offset + block->length > offset);

		// Retrieve length of overlap. May be zero if block
		// starts higher than offset.
		//
		*bl = block;
		return get_length(offset, it);
	}

	*bl = NULL;
	return 0;
}

CompressMap::con_t::const_iterator
	CompressMap::get_search_right(off_t			offset,
	                            unsigned int		level,
	                            con_t::const_iterator	it) const
{
	assert (it != m_map.end());

	off_t tmp = (*it)->offset;

	while (true)
	{
		//  offset  offset     Block
		//     |       |-----------------|
		//
		//  offset   Block    offset
		//     |---length---|    |
		//
		if (((*it)->offset <= offset) &&
		    ((*it)->offset + (*it)->length > offset) &&
		    ((*it)->level > level))
		{
			return it;
		}
		
		if ((++it == m_map.end() || (*it)->offset != tmp))
		{
			break;
		}
	}
	return m_map.end();
}

off_t CompressMap::get(off_t offset, Block **bl) const
{
	unsigned int			level;
	con_t::const_iterator		oit;
	con_t::const_iterator		it;
	con_t::const_iterator		tmp;
	Block				block (offset, UINT_MAX);

	if (m_map.empty())
	{
		*bl = NULL;
		return 0;
	}

	// Find start of the Block. If there is Block on the offset
	// it returns it, if there are more Blocks on the offset Block
	// with highest level (newest) is returned. If there is no Block
	// on the offset, Block with the highest level (newest) on
	// the nearest bigger offset is returned. Otherwise m_map.end()
	// is returned.
	//
	it = oit = m_map.lower_bound(&block);

	if (it == m_map.begin())
	{
		// There is no buffer before offset.
		//
		return get_simple(offset, it, bl);
	}

	if ((it != m_map.end()) && ((*it)->offset == offset))
	{
		level = (*it)->level;
	}
	else
		level = 0;

	// Issue 1:
	// 
	// `while` loop is there because we need to
	// find the Buffer 1 for the offset that is
	// after Buffer 3 and before Buffer 4.
	// 
	//   |-------------------|	Buffer 1
	//      |-----|	  |---|		Buffer 2, 4
	//        |----|		Buffer 3
	//
	while (true)
	{
		prev(it);

		if (it == m_map.end())
			break;
		
		tmp = get_search_right(offset, level, it);
		if (tmp != m_map.end())
		{
			// We found correct Block.
			//
			*bl = *tmp;

			// Return length of the Block.
			//
			return get_length(offset, tmp);
		}

		// Issue 1:
		//
		// Limit backward search - stop when we are on begining
		// or when we know that no Block can overlap with offset
		// we are searching for.
		//
		if (it == m_map.begin())
			break;

		if ((*it)->offset + m_length < offset)
			break;
	}

	return get_simple(offset, oit, bl);
}

void CompressMap::print() const
{
	con_t::iterator it;

	for (it = m_map.begin(); it != m_map.end(); ++it)
	{
		cout << **it << endl;
	}
	cout << "-----------------------" << endl;
}
				
void CompressMap::put(Block *bl)
{
	++m_size;

	if (bl->level == 0)
	{
		bl->level = m_size;
	}

	m_length = max(m_length, bl->length);

	m_map.insert(bl);
}

#ifdef TEST_LINEARMAP
CompressMap lm;

void get(off_t offset)
{
	off_t len;
	Block *bl;
	
	len = lm.get(offset, &bl);

	cout << "len: " << len << endl;

	if (bl)
	{
		cout << "\t" << "bl # " << bl->coffset << endl;
		cout << "\t" << "bl | offset: " << bl->offset << endl;
		cout << "\t" << "bl | length: " << bl->length << endl;
	}
	else
		cout << "\t" << "bl | NULL" << endl;
}

void add(off_t offset, size_t length, off_t id)
{
	Block *b = new (std:nothrow) Block(offset, length, id, length);
	assert (b);
	lm.put(b);
}

int main(int argc, char **argv)
{
// BIG PROBLEM:
//
//   |-------------------|	Buffer 1
//      |-----|	  |---|		Buffer 2, 4
//        |----|		Buffer 3
/*
	Block *b1 = new Block(0, 40, 1, 40);
	lm.put(b1);

	Block *b2 = new Block(5, 10, 2, 10);
	lm.put(b2);

	Block *b3 = new Block(7, 10, 3, 10);
	lm.put(b3);

	Block *b4 = new Block(20, 10, 4, 10);
	lm.put(b4);

	Block *b5 = new Block(25, 1, 5, 1);
	lm.put(b5);

	for (int i = 0; i <= 40; i++)
	{
		cout << "offset: " << i << endl;

		get(i);

		cout << "--------------------" << endl;
	}
*/
/*	// |-----------------|		Block 6
	//   |-|			Block 7
	//         |-|			Block 8
	//   |-|			Block 9
	//   
	Block *b6 = new Block(0, 26, 6, 26);
	lm.put(b6);
	
	Block *b7 = new Block(2, 1, 7, 1);
	lm.put(b7);
	
	Block *b8 = new Block(5, 1, 8, 1);
	lm.put(b8);
	
	Block *b9 = new Block(2, 1, 9, 1);
	lm.put(b9);
	
	for (int i = 0; i <= 26; i++)
	{
		cout << "offset: " << i << endl;

		get(i);

		cout << "--------------------" << endl;
	}
*/
	add(0, 5, 1);
//	add(2, 1, 2);
	add(0, 4, 3);
//	add(2, 1, 4);

	for (int i = 0; i <= 5; i++)
	{
		cout << "offset: " << i << endl;

		get(i);

		cout << "--------------------" << endl;
	}

	return 0;
}
#endif

