#include "FileRememberTimes.hpp"
#include "FileRawCompressed.hpp"
#include "TransformTable.hpp"
#include "File.hpp"

#include <algorithm>
#include <cassert>
#include <errno.h>
#include <rlog/rlog.h>
#include <cstdlib>
#include <cstring>
#include <strings.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

using namespace std;

extern TransformTable *g_TransformTable;

// Sets the type of transformation applied to index.
// We use only lzo compression to compress index for now.
// 
// Improvement?
// Have a policy that index is transformed with
// the same type of transformation as the data.
// 
FileRawCompressed::FileRawCompressed(const FileHeader &fh, off_t length) :
	m_fd (-1),
	m_fh (fh),
	m_lm (g_TransformTable->getTransform(3)),
	m_length (length),
	m_store (false)
{
	assert ((m_length == 0) || (m_length >= FileHeader::HeaderSize));

	// Limit m_length at least to FileHeader::HeaderSize - this
	// is for new files to reserve space for header...
	//
	if (m_length == 0)
	{
		m_length = FileHeader::HeaderSize;

		// Always store index and header for new
		// files...
		//
		m_store = true;
	}
}

FileRawCompressed::~FileRawCompressed()
{
	if (m_fd != -1)
	{
		close();
	}
}

int FileRawCompressed::open(int fd)
{
	assert (m_fd == -1);

	m_fd = fd;

	// Assign transformation object according to the
	// type of the transformation stored in the
	// header of the file.
	//
	m_transform = g_TransformTable->getTransform(m_fh.type);

	if (m_length == FileHeader::HeaderSize)
	{
		return m_fd;
	}
	
	// File is non-empty, so there must be index
	// there.
	// 
	assert (m_fh.index != 0);
	
	off_t r = m_lm.restore(m_fd, m_fh.index);

	if (r == -1)
	{
		// There should be index, but it cannot
		// be restored => I/O error?
		//
		m_fd = -1;
	}
	else
	{
		if (r == m_length)
		{
			// Index is stored at the end of the file.
			// 
			off_t index_length = r - m_fh.index;

			m_length -= index_length;
		}
		
	}

	return m_fd;
}

int FileRawCompressed::store(int fd)
{
	off_t r;

	FileRememberTimes frt(fd);

	// Append new index to the end of the file
	// => forget the old one...
	//
	// Set a new position of the index in the FileHeader.
	// 
	m_fh.index = m_length;

	// Store index to a file and update raw length of
	// the file.
	//
	r = m_lm.store(fd, m_fh.index);
	switch (r) {
	case -1:
		return -1;
	case  0:
		r = m_length;
		m_fh.index = 0;
	}
	
	m_length = r;
rDebug("m_fh.index = %lld", m_fh.index);	
	// Sync the header, header is always at the begining
	// of the file.
	//
	r = m_fh.store(fd, 0);
	if (r == -1)
		return -1;

	return 0;
}

int FileRawCompressed::store(const char *name)
{
	int	fd;
	off_t	r = 0;
	
	fd = File::open(name);
	if (fd < 0)
		return -1;
	
	r = store(fd);
	
	::close(fd);

	return r;
}

int FileRawCompressed::close()
{
	off_t r = 0;
	
	if (m_fd == -1)
	{
		return 0;
	}

	if (m_store)
	{
		// File was changed, write index to the file.
		//
		r = store(m_fd);
	}

	m_fd = -1;

	return r;
}

ssize_t FileRawCompressed::read(char *buf, size_t size, off_t offset)
{
	Block	*bl;
	size_t	 osize = size;
	off_t	 len;
	off_t	 r;

	while (size > 0)
	{
		// Look-up the CompressMap and find proper Block.
		//
		len = m_lm.get(offset, &bl);

		if (len)
		{
			// Block covers the offset, we can read len bytes
			// from it's de-compressed stream...
			//
			r = m_transform->restore(m_fd, bl, offset, buf,
			                         min((off_t) (size), len));
			if (r == -1)
				return -1;

			buf += r;
			offset += r;
			size -= r;
		}
		else
		{
			// Block doesn't exists on offset.
			//
			if (bl)
			{
				// There is Block on the bigger offset there,
				// fill the gap with zeroes...
				//
				// Compute the size of the gap.
				// 
				r = min(bl->offset - offset, (off_t) (size));

				memset(buf, 0, r);

				buf += r;
				offset += r;
				size -= r;
			}
			else
			{
				// There is no more data.
				//
				break;
			}
		}
	}

	// Clear the rest of the buffer.
	//
	memset(buf, 0, size);

	return osize - size;
}

ssize_t FileRawCompressed::write(const char *buf, size_t size, off_t offset)
{
	if (m_fd == -1)
	{
		errno = -EBADF;
		return -1;
	}
//	assert (m_fd != -1);

	m_store = true;

	// Append a new Block to the file.
	//
	Block *bl = m_transform->store(m_fd, offset, m_length, buf, size);
	if (!bl)
		return -1;
	
	m_lm.put(bl);
	
	// Update raw length of the file.
	// 
	m_length += bl->clength;

	// Update length of the file.
	//
	if (m_fh.size < offset + size)
	{
		m_fh.size = offset + size;
	}

	return size;
}

int FileRawCompressed::truncate(const char *name, off_t size)
{
	int r = 0;
	
	m_fh.size = size;

	m_lm.truncate(size);

	// Truncate to zero lenght is the only what
	// we can implement easily.
	//
	if (size == 0)
	{
		m_length = FileHeader::HeaderSize;

		r = ::truncate(name, m_length);
	}

	if (m_fd == -1)
	{
		// Open was not called, there is nobody who would
		// call close and thus synchronize header and index...
		//
		r = store(name);
		if (r == -1)
		{
			int tmp = errno;
			rError("FileRawCompressed::truncate Cannot write"
			       "index to file '%s', error: %s",
			       name, strerror(errno));
			errno = tmp;
		}
	}
	else
		m_store = true;

	return r;
}

int FileRawCompressed::getattr(const char *name, struct stat *st)
{
	int r;

	r = lstat(name, st);

	if ((r == 0) && (!S_ISLNK(st->st_mode)))
	{
		st->st_size = m_fh.size;
	}
	
	return r;
}

