#include <algorithm>
#include <iostream>
#include <sstream>
#include <rlog/rlog.h>
#include <errno.h>
#include <cstring>

#include <assert.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "Memory.hpp"
#include "MemoryMap.hpp"

Memory::Memory(const struct stat *st) :
	Parent (st),
	m_ParentFileSize (-1),
	m_FileSize (-1)
{
}

Memory::~Memory()
{
	assert (m_MemoryMap.empty());
}

int Memory::release(const char *name)
{
	int rm;
	int rr;

	rm = merge();
	rr = Parent::release(name);

	return rm ? rm : rr;
}

int Memory::unlink(const char *name)
{
	rDebug("Memory::unlink(%s)", m_FileName.c_str());

	// File is beeing deleted, release all memory
	// buffers allocated so far and set m_FileSize
	// to default (zero) because the object remains in
	// memory and when a new node with the same inode is
	// created it will reuse this object again...
	// 
	m_MemoryMap.truncate(0);
	m_ParentFileSize = m_FileSize = 0;

	return Parent::unlink(name);
}

int Memory::truncate(const char *name, off_t size)
{
	rDebug("Memory::truncate(%s) size: %lld", m_FileName.c_str(),
			(long long int) size);

	m_MemoryMap.truncate(size);
	m_ParentFileSize = m_FileSize = size;
	
	return Parent::truncate(name, size);
}

int Memory::getattr(const char *name, struct stat *st)
{
	int r;

	r = Parent::getattr(name, st);

	if ((r == 0) && (m_FileSize != -1))
	{
		// Override real size of the file with the internal
		// state
		// 
		st->st_size = m_FileSize;
	}

	return r;
}

int Memory::write_block(bool force)
{
	char	*buf;
	size_t	 size;
	ssize_t	 len;
	off_t	 offset;

	if (m_MemoryMap.erase(&offset, &buf, &size, force) == true)
	{
		len = Parent::write(buf, size, offset);

		delete[] buf;
	
		if (len == -1)
			return len;

		m_ParentFileSize = max(m_ParentFileSize, (off_t) (offset + size));

		return 1;
	}
	return 0;
}

ssize_t Memory::write(const char *buf, size_t size, off_t offset)
{
	int r;
	
	rDebug("Memory::write(%s) | m_FileSize: %lld, offset: %lld, size: %d",
			m_FileName.c_str(), (long long int) m_FileSize,
			(long long int) offset, (unsigned int) size);

	// Store buffer to memory in MemoryMap.
	//
	if (m_MemoryMap.put(buf, size, offset) == -1)
		return -1;

	m_FileSize = max(m_FileSize, (off_t) (offset + size));

	// Try to write a block to disk if appropriate.
	// 
	r = write_block(false);
	if (r == -1)
		return r;
	
	return size;
}

ssize_t Memory::read(char *buf, size_t size, off_t offset)
{
	char	*block;
	size_t	 block_size;
	size_t	 osize = size;
	size_t	 off;
	ssize_t	 len;
	off_t	 block_offset;

	rDebug("Memory::read(%s) | m_FileSize: %lld, offset: %lld, size: %d",
			m_FileName.c_str(), (long long int) m_FileSize,
			(long long int) offset, (unsigned int) size);

	// Allow to read up to m_FileSize.
	// 
	if ((offset + size) > m_FileSize)
	{
		if (offset > m_FileSize)
		{
			osize = size = 0;
		}
		else
			osize = size = m_FileSize - offset;
	}

	while (size > 0)
	{
		block_offset = m_MemoryMap.get(offset, &block, &block_size);

		if (block_offset == -1)
		{
			block_offset = offset + size;
		}
		
		if (block_offset > offset)
		{
			len = min(block_offset - offset, (off_t) size);
			
			if (offset > m_ParentFileSize)
			{
				// There is no memory buffer there and
				// real file is smaller than requested offset!
				//
				memset(buf, 0, len);
			}
			else
			{
				len = Parent::read(buf, len, offset);
			}
		}
		else
		{
			off = static_cast<size_t> (offset - block_offset);
			len = min(size, block_size - off);
			memcpy(buf, block + off, len);
		}

		if (len == -1)
		{
			if (osize != size)
			{
				break;
			}
			return len;
		}
	
		if (len == 0)
			break;

		buf    += len;
		offset += len;
		size   -= len;
	}

	return osize - size;
}

int Memory::open(const char *name, int flags)
{
	int		r;
	struct stat	st;

	if (m_ParentFileSize == -1)
	{
		r = Parent::getattr(name, &st);
		assert (r == 0);

		m_ParentFileSize = m_FileSize = st.st_size;
	}

	return Parent::open(name, flags);
}

/**
 * Merge memory map with content in the file on the disk.
 *
 */
int Memory::merge()
{
	int r = 1;
	
	if (m_MemoryMap.empty())
		return 0;

	Parent::truncate(m_FileName.c_str(), m_FileSize);

	while (r == 1)
	{
		r = write_block(true);
		if (r == -1)
		{
			int tmp = errno;
			rError("Memory::Merge Cannot write"
			       "data to file '%s', error: %s",
				m_FileName.c_str(), strerror(errno));
			errno = tmp;

			// Error happend, don't try to continue. But
			// don't forget to release allocated memory...
			//
			m_MemoryMap.truncate(0);
			
			return -1;
		}
	}

	return 0;
}

