#include "MemoryMap.hpp"

#include <errno.h>

#include <cassert>
#include <rlog/rlog.h>

extern size_t g_BufferedMemorySize;

MemoryMap::MemoryMap()
{
}

MemoryMap::~MemoryMap()
{
}

int MemoryMap::put(const char *buf, size_t size, off_t offset)
{
	char	*tmp;
	off_t	 tmp_offset;
	size_t	 tmp_size;
	size_t	 len;
	size_t	 off;

	assert (size > 0);

	while (size > 0)
	{
		// Check if new buffer overlaps some existing
		// Buffers already there...
		//
		tmp_offset = get(offset, &tmp, &tmp_size);
		
		if (tmp_offset == -1)
		{
			// There is no Buffer there. Cheat and
			// set tmp_offset so next 'if' creates
			// Buffer with the 'size'.
			// 
			tmp_offset = offset + size;
		}
		
		if (tmp_offset > offset)
		{
			// Create Buffer for offsets between
			// offset to tmp_offset and remember it.
			//
			len = min(tmp_offset - offset, (off_t) size);
			Buffer *buffer = new (std::nothrow) Buffer(buf, len);
			if (!buffer)
			{
				rError("No memory to allocate block of %d bytes",
						len);

				errno = -ENOMEM;
				return -1;
			}
			m_map[offset] = buffer;
		}
		else
		{
			// There is already Buffer on this offset or
			// before (with the lenght that overlaps offset
			// we are looking for).
			// 
			// tmp_offset
			// |---tmp_size---|(----------|)
			//              offset
			//              |---size---|

			// This is difference between offset we search and
			// offset we got...
			// 
			off = (size_t) (offset - tmp_offset);
			len = min(size, tmp_size - off);
			memcpy(tmp + off, buf, len);
		}
		offset += len;
		buf    += len;
		size   -= len;
	}
	return 0;
}

MemoryMap::con_t::const_iterator MemoryMap::get(off_t offset) const
{
	con_t::const_iterator it;
	con_t::const_iterator itp;

	if (m_map.empty())
		return m_map.end();
	
	it = itp = m_map.upper_bound(offset);
	
	--itp;

	if (itp != m_map.end())
	{
		if (itp->first + itp->second->size > offset)
		{
			return itp;
		}
	}
	
	return it;
}

/**
 * Finds Buffer thats start or its start is higher than
 * offset. Returns it's parameter.
 */ 
off_t MemoryMap::get(off_t offset, char **buf, size_t *size) const
{
	con_t::const_iterator	 it;
	Buffer			*buffer;

	it = get(offset);

	if (it == m_map.end())
		return -1;

	buffer = it->second;
	
	*buf = buffer->buf;
	*size = buffer->size;

	return it->first;
}

void MemoryMap::truncate(off_t size)
{
	con_t::iterator it;
	con_t::iterator itp;

	if (m_map.empty())
		return;

	it = itp = m_map.upper_bound(size);
	
	--itp;

	if ((itp != m_map.end()) &&
	    (itp->first + itp->second->size > size))
	{
		// Truncate this Buffer.
		//
		off_t off = (itp->first + itp->second->size) - size;
		
		Buffer *buffer = new (std::nothrow) Buffer(itp->second->buf,
					    itp->second->size - off);

		delete itp->second;

		// Replace old Buffer with new truncated one.
		//
		itp->second = buffer;
	}
	
	// Delete all Buffers after this iterator
	// 
	while (it != m_map.end())
	{
		delete it->second;
		m_map.erase(it++);
	}
}

size_t MemoryMap::find_total_length (con_t::const_iterator it) const
{
	size_t	size = 0;
	off_t	from = it->first;

	for (; it != m_map.end() && it->first == from; it++)
	{
		size += it->second->size;
		from += it->second->size;
	}
	
	return size;
}

void MemoryMap::copy_all	(con_t::iterator it, char *buf, ssize_t len)
{
	assert (it != m_map.end());

	off_t					 from = it->first;
	Buffer					*buffer;
	
	while (it != m_map.end() && it->first == from)
	{
		buffer = it->second;
		
		memcpy(buf, buffer->buf, buffer->size);
		buf  += buffer->size;
		from += buffer->size;
		len  -= buffer->size;

		delete buffer;
		m_map.erase(it++);
	}
	assert (len == 0);
}

bool MemoryMap::erase(off_t *offset, char **buf, size_t *size, bool force)
{
	size_t		len;
	con_t::iterator	it;

	*offset = 0;
	
	while ((it = m_map.lower_bound(*offset)) != m_map.end())
	{
		*offset = it->first;

		len = find_total_length(it);
		assert (len > 0);
		
		if ((force == true) || (len >= g_BufferedMemorySize))
		{
			*size = len;
			*buf = new (std::nothrow) char[len];
			assert (*buf);
		
			copy_all(it, *buf, len);

			return true;
		}
		
		*offset += len;
	}
	return false;
}

#ifdef TEST_PLAINMAP

int main(int argc, char **argv)
{
	MemoryMap pm(5);

	pm.put("AHOJ", 4, 0);
	pm.put("SVETE", 5, 7);
	pm.put("NNN", 3, 4);

	pm.truncate(6);

	char *b;
	size_t s;
	off_t o;
	off_t i = 0;

	while ((o = pm.get(i, &b, &s)) != -1)
	{
		if (o > i)
			cout << " ";
		else
			cout << b[i-o];
		i++;
	}
	cout << endl;

	if (pm.erase(&o, &b, &s, true))
	{
		char *ob = b;
		cout << "offset: " << o << endl;
		cout << "size: " << s << endl;
		cout << "buf: ";
		while (s-- > 0)
			cout << *b++;
		cout << endl;
		delete[] ob;
	}

	return 0;
}

#endif

